!=========================================================================
! Copyright (C) GemTalk Systems 1986-2015.  All Rights Reserved.
!
! $Id$
!
! Superclass Hierarchy:
!   Object, GsObjectInventory
!
!=========================================================================

expectvalue %String
run
^Object _newKernelSubclass: 'GsObjectInventory'
        instVarNames: #( entriesByCount entriesByBytes includeHiddenClasses)
        classVars: #()
        classInstVars: #()
        poolDictionaries: #[]
        inDictionary: Globals
        constraints: #[ ]
        instancesInvariant: false
        isModifiable: false
        reservedOop: 1137
%

removeallmethods GsObjectInventory
removeallclassmethods GsObjectInventory

! ------------------- Class methods for GsObjectInventory

category: 'Repository Analysis'
classmethod: GsObjectInventory
profileRepository

"Scans the entire repository and creates a profile by class.  Returns a new
 instance of the receiver.

 Counts the number of instances and total physical bytes occupied of each
 class in the repository with one or more instances.

 Objects which are pending garbage collection reclamation are not included
 in the result.

 IMPORTANT:  This method stays in a transaction for its entire duration and
             is therefore not recommended for use on production systems.

 Requires the #GarbageCollection privilege.

 Acquires the garbage collection lock which prevents running garbage
 collection or other scans which requires the GC lock while this method
 is running.  

 Raises error #rtErrGetGcLockFailed if the GC lock could not be acquired
 within 3 minutes.

 Aborts the current transaction.   If this session contains uncommitted
 changes to the repository, the method  signals a error: 
 #rtErrAbortWouldLoseData, to indicate that data could be lost. Otherwise
 it puts the session into auto-begin transaction mode and aborts."

  System needsCommit
    ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self _profileRepositoryAndShowHiddenClasses: true
%

category: 'Repository Analysis'
classmethod: GsObjectInventory
profileRepositoryAndSkipHiddenClasses

"Performs the same function as the #profileRepository method except that 
 statistics for GemStone private classes (e.g. LargeObjectNode,
 NscSetLeaf, etc) are not shown.  In effect, instances of private classes
 are ignored during the scan.  However the approximate physical space 
 occupied by private objects is included in the physical size of the
 private object's root object, which will always be a public class.
 For example, the approximate physical bytes consumed by a NscSetLeaf object
 will be included in the physical size reported for root object, which will be
 an IdentitySet (or a subclass of IdentitySet)."

^self _profileRepositoryAndShowHiddenClasses: false
%

category: 'Garbage Analysis'
classmethod: GsObjectInventory
profileGarbageFromFile: aFilename includeHiddenObjects: aBoolean

"Analyze the disconnected (garbage) objects in the repository using
 aFilename (an instance of String), which was produced by one of the 
 Repository>>findDisconnectedObjectsAndWriteToFile: methods.  Only objects
 which still exist in the repository are analyzed.  Does not create
 references to any object in the file.

 Scans the entire repository and creates a profile by class of the objects
 in the file.  Returns a new instance of the receiver.

 DeadNotReclaimed objects (those pending garbage collection reclamation)
 are not included in the result.

 IMPORTANT:  This method stays in a transaction for its entire duration and
             is therefore not recommended for use on production systems.

 Requires the #GarbageCollection privilege.

 Acquires the garbage collection lock which prevents running garbage
 collection or other scans which requires the GC lock while this method
 is running.  

 Raises error #rtErrGetGcLockFailed if the GC lock could not be acquired
 within 3 minutes.

 Aborts the current transaction.   If this session contains uncommitted
 changes to the repository, the method  signals a error: 
 #rtErrAbortWouldLoseData, to indicate that data could be lost. Otherwise
 it puts the session into auto-begin transaction mode and aborts."

System _hiddenSetReinit: 34 .
SystemRepository loadGcCandidatesFromFile: aFilename intoHiddenSet: 34 .
^ self _primProfileHiddenSet: 34 showHiddenClasses: aBoolean
%

category: 'Private'
classmethod: GsObjectInventory
_profileRepositoryAndShowHiddenClasses: aBoolean

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _primProfileRepositoryAndShowHiddenClasses: aBoolean
%

category: 'Private'
classmethod: GsObjectInventory
_primProfileRepositoryAndShowHiddenClasses: aBoolean

<primitive: 677>
aBoolean _validateClass: Boolean .
^ self _primitiveFailed: #_primProfileRepositoryAndShowHiddenClasses: .
%

category: 'Private'
classmethod: GsObjectInventory
_primProfileHiddenSet: hiddenSetSpecifier showHiddenClasses: aBoolean

"Scans the entire repository in page order to profile the instances in
 the given hidden set.  Should only be used for large sets of objects."

<primitive: 678>
hiddenSetSpecifier _validateClass: SmallInteger .
(hiddenSetSpecifier < 1 
  or:[hiddenSetSpecifier > (System HiddenSetSpecifiers size)])
    ifTrue: [hiddenSetSpecifier _error: #rtErrArgOutOfRange].
aBoolean _validateClass: Boolean .
^ self _primitiveFailed: #_primProfileHiddenSet:showHiddenClasses: .
%

category: 'Private'
classmethod: GsObjectInventory
_primProfileSmallHiddenSet: hiddenSetSpecifier showHiddenClasses: aBoolean

<primitive: 695>
hiddenSetSpecifier _validateClass: SmallInteger .
(hiddenSetSpecifier < 1 
  or:[hiddenSetSpecifier > (System HiddenSetSpecifiers size)])
    ifTrue: [hiddenSetSpecifier _error: #rtErrArgOutOfRange].
aBoolean _validateClass: Boolean .
^ self _primitiveFailed: #_primProfileSmallHiddenSet:showHiddenClasses: .
%

category: 'Profiling'
classmethod: GsObjectInventory
profileObjectsInHiddenSet: anInt showHiddenClasses: aBool

"Generate a profile of the objects in the given hidden set which
 exist and for which the session has permission to read.  Objects
 which do not meet these criteria are silently omitted from the result.

 This method does not alter the contents of the hidden set.

 Returns an instance of the receiver."

^ self _primProfileSmallHiddenSet: anInt showHiddenClasses: aBool
%

category: 'Profiling'
classmethod: GsObjectInventory
profileObjectsIn: aCollection

"Generate a profile of the objects in the given collection.
 The collection must not contain any special objects, otherwise
 an error will be raised.

 Returns an instance of the receiver."

| anArray |
anArray := aCollection asArray .
System _hiddenSetReinit: 26 .
System _addAll: anArray to: 26.
^ self profileObjectsInHiddenSet: 26 showHiddenClasses: true
%

! ------------------- Instance methods for GsObjectInventory
category: 'Formatting'
method: GsObjectInventory
asString

| sz result|
sz := self entriesByCount size.
result := String new.
result addAll: 'a'; addAll: self class name; addAll: ' of '; addAll: sz asString; addAll: ' classes.'.
^result
%
category: 'Reporting'
method: GsObjectInventory
byteCountReport
	^self byteCountReportDownTo: 0

%
category: 'Reporting'
method: GsObjectInventory
byteCountReportDownTo: minCountToReport

"Return a String showing the receiver in tabular form sorted by instance count."

| totalBytes totalCount result tmp lf array line space|
lf := Character lf.
space := Character space.
result := String new.
tmp := String withAll: '*** GsObjectInventory byteCountReport printed at: '.
tmp addAll: DateTime now asString.
tmp addAll: ' ***'.
result addLineWith: tmp centeredToWidth: 80 ; 
       add: lf.
result addAll: 'Hidden classes are '.
self includeHiddenClasses
  ifFalse:[ result addAll: 'not '.].
result addAll: 'included in this report.'; add: lf.
line := String new.
80 timesRepeat:[line add: $_].
line add: lf.
result addAll: line.

result 	addAll: ((String withAll: 'Class') width: 50; yourself); add: space;
		addAll: ((String withAll: 'Instances') width: -14; yourself); add: space;
		addAll: ((String withAll: 'Bytes') width: -14; yourself); add: lf.
result addAll: line.
totalBytes := 0.
totalCount := 0.
array := self entriesByBytes.
1 to: array size do:[:n| |entry|
  entry := array at: n.
  (entry byteCount < minCountToReport)
	ifTrue:[result addAll: line.
		^result ].
  result addAll: (entry theClass name asString width: 50); add: space;
		addAll: (entry instanceCount asString width: -14); add: space;
		addAll: (entry byteCount asString width: -14); add: lf.
  totalCount := totalCount + entry instanceCount.
  totalBytes := totalBytes + entry byteCount.
].
result addAll: line.
result 	addAll: ((String withAll: 'Totals') width: 50; yourself); add: space;
        addAll: (totalCount asString width: -14) ; add: space;
        addAll: (totalBytes asString width: -14); add: lf.
result addAll: line.
^result
%
category: 'Accessing'
method: GsObjectInventory
entriesByBytes

^entriesByBytes
%
category: 'Updating'
method: GsObjectInventory
entriesByBytes: newValue

entriesByBytes := newValue
%
category: 'Accessing'
method: GsObjectInventory
entriesByCount

^entriesByCount
%
category: 'Accessing'
method: GsObjectInventory
includeHiddenClasses

^includeHiddenClasses
%
category: 'Updating'
method: GsObjectInventory
entriesByCount: newValue

entriesByCount := newValue
%
category: 'Updating'
method: GsObjectInventory
includeHiddenClasses: newValue

includeHiddenClasses := newValue
%
category: 'Reporting'
method: GsObjectInventory
instanceCountReport

"Return a String showing the receiver in tabular form sorted by instance count."
^self instanceCountReportDownTo: 0
%
category: 'Reporting'
method: GsObjectInventory
instanceCountReportDownTo: minCountToReport

"Return a String showing the receiver in tabular form sorted by instance count."
| totalCount totalBytes result tmp lf array line space|
lf := Character lf.
space := Character space.
result := String new.
tmp := String withAll: '*** GsObjectInventory instanceCountReport printed at: '.
tmp addAll: DateTime now asString.
tmp addAll: ' ***'.
result addLineWith: tmp centeredToWidth: 80 ; 
       add: lf.
result addAll: 'Hidden classes are '.
self includeHiddenClasses
  ifFalse:[ result addAll: 'not '.].
result addAll: 'included in this report.'; add: lf.

line := String new.
80 timesRepeat:[line add: $_].
line add: lf.
result addAll: line.
result 	addAll: ((String withAll: 'Class') width: 50; yourself); add: space;
		addAll: ((String withAll: 'Instances') width: -14; yourself); add: space;
		addAll: ((String withAll: 'Bytes') width: -14; yourself); add: lf.
result addAll: line.
array := self entriesByCount.
totalCount := 0.
totalBytes := 0.
1 to: array size do:[:n| |entry|
  entry := array at: n.
  (entry instanceCount < minCountToReport)
	ifTrue:[	result addAll: line.
			^result
	].
  result addAll: (entry theClass name asString width: 50); add: space;
		addAll: (entry instanceCount asString width: -14); add: space;
		addAll: (entry byteCount asString width: -14); add: lf.
  totalCount := totalCount + entry instanceCount.
  totalBytes := totalBytes + entry byteCount.
].
result addAll: line.
result 	addAll: ((String withAll: 'Totals') width: 50; yourself); add: space;
        addAll: (totalCount asString width: -14) ; add: space;
        addAll: (totalBytes asString width: -14); add: lf.
result addAll: line.
^result
%

! Start request 43179

category: 'Private'
classmethod: GsObjectInventory
_primProfileRepositoryAndShowHiddenClasses: aBoolean andListInstancesInPageOrder: anIdentitySet
toFile: aString

<primitive: 697>
aBoolean _validateClass: Boolean .
anIdentitySet _validateClass: IdentitySet .
(anIdentitySet size between: 1 and: 2034)
  ifFalse:[ anIdentitySet _error: #rtErrArgOutOfRange ].
anIdentitySet do: [:each | each _validateClass: Behavior].

aString _validateClass: String .
aString trimBlanks size > 0
  ifFalse:[ aString _error: #rtErrArgOutOfRange ].

^ self _primitiveFailed: 
  #_primProfileRepositoryAndShowHiddenClasses:andListInstancesInPageOrder:toFile:
%

category: 'Repository Analysis'
classmethod: GsObjectInventory
profileRepositoryAndListInstancesInPageOrder: anArrayOfClasses toFile: aString

"Combines the functions of the following methods in a single scan of the 
 repository:

   GsObjectInventory>>profileRepository
   Repository>>listInstancesInPageOrder: toFile:

 Refer to the comments in these methods for more information.

 Requires the #GarbageCollection privilege.

 Acquires the garbage collection lock which prevents running garbage
 collection or other scans which requires the GC lock while this method
 is running.  

 Raises error #rtErrGetGcLockFailed if the GC lock could not be acquired
 within 3 minutes.

 Aborts the current transaction.   If this session contains uncommitted
 changes to the repository, the method  signals a error: 
 #rtErrAbortWouldLoseData, to indicate that data could be lost. Otherwise
 it puts the session into auto-begin transaction mode and aborts.

 Returns an Array containing 2 elements:
  [1] - An new instance of the receiver.
  [2] - An Integer indicating the number of object identifiers written 
        to the file."

| idSet |
System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

idSet := IdentitySet withAll: anArrayOfClasses .
^ self _primProfileRepositoryAndShowHiddenClasses: true 
       andListInstancesInPageOrder: idSet
       toFile: aString
%

category: 'Repository Analysis'
classmethod: GsObjectInventory
profileRepositoryAndSkipHiddenClassesAndListInstancesInPageOrder: anArrayOfClasses toFile: aString

"Combines the functions of the following methods in a single scan of the 
 repository:

   GsObjectInventory>>profileRepositoryAndSkipHiddenClasses
   Repository>>listInstancesInPageOrder: toFile:

 Refer to the comments in these methods for more information.

 Requires the #GarbageCollection privilege.

 Acquires the garbage collection lock which prevents running garbage
 collection or other scans which requires the GC lock while this method
 is running.  

 Raises error #rtErrGetGcLockFailed if the GC lock could not be acquired
 within 3 minutes.

 Aborts the current transaction.   If this session contains uncommitted
 changes to the repository, the method  signals a error: 
 #rtErrAbortWouldLoseData, to indicate that data could be lost. Otherwise
 it puts the session into auto-begin transaction mode and aborts.

 Returns an Array containing 2 elements:
  [1] - An new instance of the receiver.
  [2] - An Integer indicating the number of object identifiers written 
        to the file."

| idSet |
System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

idSet := IdentitySet withAll: anArrayOfClasses .
^ self _primProfileRepositoryAndShowHiddenClasses: false 
       andListInstancesInPageOrder: idSet
       toFile: aString
%
